#include "MHMesh.h"

//using a macro here because difficult spelling
#define setmptrel(matrixptr,coord1,coord2,value) matrixptr->operator()(coord1,coord2) = value

//same
//#define getvec3(floatptr) 

//void getvec3(float* floatptr, Vector3f target);


bool MHMesh::calcMHB()
{
	bool meshOK = true;

	mesh->add_property(index,"index");
	mesh->add_property(vtarea,"vtarea");


	//add indices to vertices
	//=======================
	cout << "MHMESH: Indexing vertices... ";
	int idx=0;

	Mesh::VertexIter vertexiterator = mesh->vertices_begin();
	for(vertexiterator; vertexiterator != mesh->vertices_end(); ++vertexiterator)
	{
		mesh->property(index,vertexiterator) = idx;					//simply number the vertices in their current order
		++idx;
	}

	numberofvertices = mesh->n_vertices();
	numEigVecs = numberofvertices;
	if(numberofvertices<numEigVecs)
	{
		cout<<"ERROR: The number of vertices is less than desired frequencies"<<endl;
		throw;
	}
	cout << "done." << endl;




	//add voronoi area to vertices
	//============================
	cout << "MHMESH: Computing voronoi areas... ";
	vertexiterator = mesh->vertices_begin();
	for(vertexiterator; vertexiterator != mesh->vertices_end(); ++vertexiterator)
	{
		float currentarea = 0;		//add each triangle's area to this sum

		float* currentdata = mesh->point(vertexiterator).data();				//current vertex' position
		Vector3f currentpoint(currentdata[0],currentdata[1],currentdata[2]);

		bool lastneighbor = false;

		Mesh::VertexVertexIter neighboriterator = mesh->vv_begin(vertexiterator); //- 1 ring neighbour iterator
		int neighborNo = 0; //needed for break condition

		for( ; ; )	
		{
			//last neighbor reached
			if(neighboriterator == mesh->vv_end(vertexiterator) && neighborNo > 0){
				break;
			}

			float* previousdata = mesh->point(neighboriterator).data();					// current neighbor's position
			Vector3f previouspoint(previousdata[0],previousdata[1],previousdata[2]);

			++neighboriterator; //for each neighbor take the next neighbor, but for the last one take the first one and break
			neighborNo++; 


			float* nextdata = mesh->point(neighboriterator).data();						//next neighbor's position
			Vector3f nextpoint(nextdata[0],nextdata[1],nextdata[2]);

			//calculate centers of edges and of the face area
			Vector3f previouscenter = (currentpoint+previouspoint)/2;
			Vector3f nextcenter		= (currentpoint+nextpoint)/2;
			Vector3f facecenter		= (currentpoint+previouspoint+nextpoint)/3;

			//calculate their lengths
			float	previouslength	= (currentpoint-previouscenter).norm();
			float	nextlength		= (currentpoint-nextcenter).norm();
			float	currentface		= (currentpoint-facecenter).norm();
			float	previousface	= (previouscenter-facecenter).norm();
			float	nextface		= (nextcenter-facecenter).norm();

			//calculate voronoi area in this triangle, in two triangular halves
			//area = sqrt(s*(s-a)*(s-b)*(s-c))
			//s = circumference/2
			float firsts		= (previouslength + previousface + currentface)/2;
			float seconds		= (nextlength + nextface + currentface)/2;
			float firstarea		= sqrt(firsts*(firsts-previouslength)*(firsts-previousface)*(firsts-currentface));
			float secondarea	= sqrt(seconds*(seconds-nextlength)*(seconds-nextface)*(seconds-currentface));


			float triarea		= firstarea+secondarea;

			currentarea += triarea;

		}

		mesh->property(vtarea,vertexiterator) = currentarea;

	}
	cout << "done." << endl;




	//calculate "triangle" operator
	cout << "MHMESH: Computing Laplace operator... ";
	//=============================
	MatrixXd* tri = new MatrixXd(numberofvertices,numberofvertices);

	//fill triangle with zeros
	tri->setZero();

	//go through all vertices and calculate their coefficients...
	vertexiterator = mesh->vertices_begin();
	for(vertexiterator; vertexiterator != mesh->vertices_end(); ++vertexiterator)
	{
		float diagsum = 0;	//sum for the diagonal entry

		int i = mesh->property(index,vertexiterator);	//index of this vertex 
		float firstarea = mesh->property(vtarea,vertexiterator);	//area of this vertex

		bool lastneighbor = false; //when to break the loop

		float* currentdata = mesh->point(vertexiterator).data();				//current vertex' position
		Vector3f currentpoint(currentdata[0],currentdata[1],currentdata[2]);

		//for all its neighbors 
		Mesh::VertexVertexIter neighboriterator = mesh->vv_begin(vertexiterator); 

		int neighborNo=0;
		for( ; ; )
		{
			if(neighboriterator == mesh->vv_end(vertexiterator) && neighborNo > 0){
				break;
			}

			int j = mesh->property(index,neighboriterator);		//index of neighbor
			float secondarea = mesh->property(vtarea,neighboriterator);	//area of neighbor

			float* data = mesh->point(neighboriterator).data();
			Vector3f currentneighbor(data[0],data[1],data[2]);

			Vector3f nextneighbor;
			Vector3f previousneighbor;


			//if at first neighbor, use last neighbor as previous
			//if at last neighbor, use first neghgbor as next
			//in between: proceed sequentially
			if(neighboriterator == mesh->vv_begin(vertexiterator))
			{
				++neighboriterator;
				float* nextdata = mesh->point(neighboriterator).data();
				nextneighbor = Vector3f(nextdata[0],nextdata[1],nextdata[2]);

				float* prevdata = mesh->point(--mesh->vv_end(vertexiterator)).data();
				previousneighbor = Vector3f(prevdata[0],prevdata[1],prevdata[2]);

				if(currentneighbor[0]==previousneighbor[0] && currentneighbor[1]==previousneighbor[1] && currentneighbor[2]==previousneighbor[2])
					cout << "Vertex" << i << ", neighbor 0 and " << neighborNo << ": same coordinates."<< endl;
			}
			else
			{
				--neighboriterator;
				float* prevdata = mesh->point(neighboriterator).data();
				previousneighbor=Vector3f(prevdata[0],prevdata[1],prevdata[2]);

				++neighboriterator;
				++neighboriterator;
				float* nextdata = mesh->point(neighboriterator).data();
				nextneighbor=Vector3f(nextdata[0],nextdata[1],nextdata[2]);
			}



			//get the angles
			//first case
			Vector3f prevcur		= (currentneighbor  - previousneighbor ).normalized();
			Vector3f prevcent		= (currentpoint  - previousneighbor ).normalized();
			float firstangle		= acos(prevcur.dot(prevcent));

			//second case
			Vector3f nextcur		= (currentneighbor  - nextneighbor ).normalized();
			Vector3f nextcent		= (currentpoint  - nextneighbor ).normalized();
			float secondangle		= acos(nextcur.dot(nextcent));

			float firstcot			= 1/tan(firstangle);
			float secondcot			= 1/tan(secondangle);



			float coefficient		= (-1) * (firstcot + secondcot) / sqrt(firstarea*secondarea);

			//handle corrupt meshes
			int cl = _fpclass(coefficient);
			if(!(cl==_FPCLASS_NN || cl==_FPCLASS_PN)){
				meshOK=false;
				coefficient = 0;	//set corrupt value to 0.. the thug life chose me.
			}

			setmptrel(tri,i,j,coefficient);

			diagsum += coefficient;


			neighborNo++;

		}	//for neighbors


		setmptrel(tri,i,i,(-1)*diagsum);		//case i=j

	}	//for vertices
	cout << "done." << endl;


	cout << "MHMESH: Computing eigenvectors... ";
	EigenSolver<MatrixXd> *solver = new EigenSolver<MatrixXd>(*tri,true);

	//get the $numEigVecs eigenvectors with smallest eigenvalues
	mhb = MatrixXd(numberofvertices,numEigVecs);
	eigenvalues = VectorXd(numEigVecs);

	MatrixXd eigVecsTmp = solver->eigenvectors().real();
	VectorXd eigValsTmp = solver->eigenvalues().real();
	vector<bool> taken(numberofvertices,false);

	for(int i=0; i<numEigVecs; i++){
		double smallestAbs = 3.40282e+038;	//double max
		int smallestIdx = 0;
		for(int j=0; j<numberofvertices; j++){
			if(!taken.at(j)){
				if(abs(eigValsTmp(j))<smallestAbs){
					smallestAbs = abs(eigValsTmp(j));
					smallestIdx = j;
				}
			}
		}

		mhb.col(i)=eigVecsTmp.col(smallestIdx);
		eigenvalues(i)=eigValsTmp(smallestIdx);
		taken.at(smallestIdx)=true;
	}

	delete solver;

	delete tri;

	cout << "done." << endl;

	mhbcalced = true;
	return meshOK;
}


void MHMesh::calcMHT()
{
	if(!mhbcalced)
	{
		throw "mhb has not yet been calculated";
		return;
	}

	//get vertex components as vectors
	Mesh::VertexIter vertexiterator = mesh->vertices_begin();
	VectorXd currentX(numberofvertices);
	VectorXd currentY(numberofvertices);
	VectorXd currentZ(numberofvertices);

	for(vertexiterator; vertexiterator != mesh->vertices_end(); ++vertexiterator)
	{
		float* currentdata = mesh->point(vertexiterator).data();				
		int idx = mesh->property(index,vertexiterator);

		currentX(idx) = currentdata[0];
		currentY(idx) = currentdata[1];
		currentZ(idx) = currentdata[2];
	}

	int numHarmonics = mhb.cols();

	VectorXd t_mhtX(numHarmonics);
	VectorXd t_mhtY(numHarmonics);
	VectorXd t_mhtZ(numHarmonics);

	for(int i=0;i<numHarmonics;i++)
	{
		VectorXd currentHarmonic = mhb.col(i);

		//multiply  vertices transpose (row)*harmonic (column) ; result is 1x1 matrix float value()
		t_mhtX(i) = (currentX.transpose()* currentHarmonic ).value();
		t_mhtY(i) = (currentY.transpose()*currentHarmonic).value();
		t_mhtZ(i) = (currentZ.transpose()*currentHarmonic).value();
	}

	mhtX = t_mhtX;
	mhtY = t_mhtY;
	mhtZ = t_mhtZ;

	mhtcalced=true;
}

bool MHMesh::reconstruct(int numEigVecs)
{
	if(!mhtcalced)
		return false;
	clock_t t=clock();

	int numVertices = mhb.rows();

	VectorXd t_reconstructedX(numVertices);
	VectorXd t_reconstructedY(numVertices);
	VectorXd t_reconstructedZ(numVertices);

	MatrixXd currentMhb = mhb.block(0,0,numVertices,numEigVecs);

	VectorXd currentMhtX = mhtX.head(numEigVecs);
	VectorXd currentMhtY = mhtY.head(numEigVecs);
	VectorXd currentMhtZ = mhtZ.head(numEigVecs);

	VectorXd currentFilter = filter.head(numEigVecs);


	for(int i=0;i<numVertices;i++)
	{
		VectorXd currentInverseHarmonic = currentMhb.row(i);

		t_reconstructedX(i) = (currentInverseHarmonic.transpose()* currentMhtX.cwiseProduct(currentFilter) ).value();
		t_reconstructedY(i) = (currentInverseHarmonic.transpose()* currentMhtY.cwiseProduct(currentFilter) ).value();
		t_reconstructedZ(i) = (currentInverseHarmonic.transpose()* currentMhtZ.cwiseProduct(currentFilter) ).value();


	}

	reconstructedX = t_reconstructedX;
	reconstructedY = t_reconstructedY;
	reconstructedZ = t_reconstructedZ;

	return true;
}

float *MHMesh::getVert(int ind){

	float ret[] = {reconstructedX[ind],
		reconstructedY[ind],
		reconstructedZ[ind]};
	return ret;
}

float *MHMesh::getPositions(){
	float *positions = new float[numberofvertices*3];

	for(int i=0; i<numberofvertices; i++){
		positions[i*3] = reconstructedX[i];
		positions[i*3+1] = reconstructedY[i];
		positions[i*3+2] = reconstructedZ[i];
	}

	return positions;
}


//just for debug purposes
void MHMesh::writeToFiles(){
	ofstream _mhb;
	_mhb.open("_mhb.txt");
	_mhb.precision(10);
	_mhb<<mhb;
	_mhb.close();

	ofstream val;
	val.open("_eigenvalues.txt");
	val.precision(10);
	val<<eigenvalues;
	val.close();

	ofstream pointcloud;
	ofstream comparison;
	comparison.open("_comparison.txt");
	pointcloud.open("_pointcloud.txt");
	Mesh::VertexIter vertexiterator = mesh->vertices_begin();
	int numberofvertices = mesh->n_vertices();
	for(vertexiterator; vertexiterator != mesh->vertices_end(); ++vertexiterator)
	{
		float* currentdata = mesh->point(vertexiterator).data();				
		int idx = mesh->property(index,vertexiterator);
		comparison<<"original XYZ: "<< currentdata[0]<<","<<currentdata[1]<<","<<currentdata[2]<<" - reconstructed: "<<reconstructedX(idx)<<","<<reconstructedY(idx)<<","<<reconstructedZ(idx)<<endl;
		pointcloud << reconstructedX(idx)<<","<<reconstructedY(idx)<<","<<reconstructedZ(idx)<<endl;
	}
	comparison.close();
	pointcloud.close();

	ofstream mht;
	mht.open("_mht.txt");
	for(int i=0; i<numEigVecs; i++)
	{
		mht << mhtX(i)<<","<<mhtY(i)<<","<<mhtZ(i)<<endl;
	}
	mht.close();
}