/**
@file MeshViewerWidget.h
@brief code taken from OpenMesh and adopted to our needs
*/

#ifndef OPENMESHAPPS_VIEWERWIDGET_HH
#define OPENMESHAPPS_VIEWERWIDGET_HH

//== INCLUDES =================================================================
#include <QWidget>
#include <QString>
#include <QFileDialog>
#include <QPushButton>
#include <QColor>
#include <QColorDialog>
#include <OpenMesh/Tools/Utils/getopt.h>
#include <OpenMesh/Tools/Utils/Timer.hh>
#include "MeshViewerWidgetT.h"
#include <OpenMesh/Core/Mesh/TriMesh_ArrayKernelT.hh>
#include <math.h>

#include <QVBoxLayout>
#include <QHBoxLayout>
#include <qpushbutton.h>



//== CLASS DEFINITION =========================================================

using namespace OpenMesh;  
using namespace OpenMesh::Attributes;



//== CLASS DEFINITION =========================================================

//class MeshViewerWidget : public MeshViewerWidgetT<MyMesh>
class MeshViewerWidget : public MeshViewerWidgetT<OpenMesh::TriMesh_ArrayKernelT<OpenMesh::DefaultTraits>>
{
    Q_OBJECT
public:
    /// default constructor
    //MeshViewerWidget(QWidget* parent=0) : MeshViewerWidgetT<MyMesh>(parent)
	MeshViewerWidget(QWidget* parent=0) : MeshViewerWidgetT<OpenMesh::TriMesh_ArrayKernelT<OpenMesh::DefaultTraits>>(parent)
    {
		mhMesh = new MHMesh();
		mhMesh->mesh = &mesh_;
		t=clock();
	}
    OpenMesh::IO::Options& options() { return _options; }
    const OpenMesh::IO::Options& options() const { return _options; }
    void setOptions(const OpenMesh::IO::Options& opts) { _options = opts; }

    void open_mesh_gui(QString fname)
    {
        OpenMesh::Utils::Timer t;
        t.start();
		LoadMeshErr err = open_mesh(fname.toLocal8Bit(), _options);
        
		if ( fname.isEmpty() || err==UNKNOWN)
        {
			QString msg = "Cannot read mesh from file:\n '";
			msg += fname;
			msg += "'";
            QMessageBox::critical( NULL, windowTitle(), msg);
        }
		if(err==FEW_VERTS){
			QString msg = "Your mesh must have at least 64 vertices!";
            QMessageBox::information( NULL, windowTitle(), msg);
		}
		if(err==MANY_VERTS){
			QString msg = "Computing the manifold harmonics of a mesh with n vertices includes ";
			msg+="computing the eigenvectors of an nxn-matrix, which has a complexity of O(n^3). ";
			msg+="So, for your own good, please try to load a mesh with less than 1000 vertices.";
            QMessageBox::information( NULL, windowTitle(), msg);
		}
		if(err==CORRUPT){
			QString msg = "Your mesh should be manifold and closed in order to look good after reconstruction.";
            QMessageBox::information( NULL, windowTitle(), msg);
		}
        
		t.stop();
		std::cout << "Loaded mesh in ~" << t.as_string() << std::endl;
    }


    void open_audio_gui(QString fname)
    {
        if ( fname.isEmpty() || !open_audio( fname.toLocal8Bit() ) )
        {
            QString msg = "Cannot load audio file from file:\n '";
            msg += fname;
            QMessageBox::warning( NULL, windowTitle(), msg );
        }
    }
	

public slots:
    void query_open_mesh_file() {
        QString fileName = QFileDialog::getOpenFileName(this,
            tr("Open mesh file"),
            tr(""),
            tr("OBJ Files (*.obj);;"
            "OFF Files (*.off);;"
            "STL Files (*.stl);;"
            "All Files (*)"));
        if (!fileName.isEmpty())
            open_mesh_gui(fileName);
    }
    void query_open_audio_file() {
        QString fileName = QFileDialog::getOpenFileName(this,
            tr("Open audio file"),
            tr(""),
            tr("Mp3 Files (*.mp3);;"
            "All Files (*)"));
        if (!fileName.isEmpty())
            open_audio_gui(fileName);
    }
	void changeFreqs(int freqs)
	{
		int newfreqs = (int)pow((double)2,freqs);
		if(newfreqs > mhMesh->mesh->n_vertices()){
			return;
		}
		Config::numFreqs = newfreqs;
		audio.setFreqs(Config::numFreqs);
	}
	void changeSmoothing(int s)
	{
		Config::smoothing = s;
	}

	
	void playPause(){
		audio.playPause();
	}
	void rePlay(){
		audio.rePlay();
	}

	void eqLow(int val){
		audio.updateEQ(LOW, val);
	}
	void eqMid(int val){
		audio.updateEQ(MID, val);
	}
	void eqHigh(int val){
		audio.updateEQ(HIGH, val);
	}
	void query_change_color(){
		QColor c = QColorDialog::getColor();
		float r = (float)c.red()/255;
		float g = (float)c.green()/255;
		float b = (float)c.blue()/255;
		change_color(r,g,b);
	}


private:
    OpenMesh::IO::Options _options;
};


#endif
