/** 
@file audio.h 
@brief defines class Audio and enum EQType
*/ 

#ifndef AUDIO
#define AUDIO

#include <fmod.hpp>
#include "common.h"
#include "common_transform.h"
#include "config.h"


/**
@enum EQType
@brief Specifies which parameter of the equalizer function should be updated.
*/
enum EQType{
	LOW,		
	MID,
	HIGH,
	FREQS};

/**
@brief responsible for all audio signal related tasks.
*/
class Audio{

public:
	/// default constructor
	Audio();
	/// destructor
	~Audio();

	/**	Loads song from specified source and starts playback.
	@param path path to an audio file
	*/
	bool loadSong(const char *path);
	
	/**	Restarts song from beginning.
	*/
	void rePlay();
	
	/**	Pause/unpause song.
	*/
	bool playPause();

	/**	Returns a filter which can be applied to the MHT.
	The filter is composed of the frequency spectrum of the currently playing audio file
	and a weight function, which the user can manipulate (see updateEQ()).
	@param filter a buffer to which the filter will be written.
	*/
	void getFilter(VectorXd *filter);

	/**	Sets the number of frequency bands that should be extracted from the audio stream/delivered as filter.
	@param _numFreqs new number of frequency bands
	*/
	void setFreqs(int _numFreqs) {
		numFreqs = _numFreqs;
		updateEQ(FREQS,0);
	}

	/**	Updates the equalizer function.
	The equalizer function is vector of numHarmonics weights.
	In [0, numFreqs/4-1] (first fourth) it is a linear interpolation between low/2 and mid*2.
	In [numFreqs/4, numFreqs-1] (rest) we interpolate between mid*2 and high*10.
	low, mid and high must range between 0 and 2.
	@param t type of parameter that should be updated
	@param val new value for that parameter (must range between 0 and 200, has no effect if t=FREQS)
	*/
	void updateEQ(EQType t, int val);

protected:
	bool ready;					/**< indicates if a song has already been loaded */ 
	float *left;				/**< buffer for frequency spectrum of left channel */ 
	float *right;				/**< buffer for frequency spectrum of right channel */
	float *eq;					/**< storage of equalizer function */
	float low;					/**< parameter that influences weight of low frequencies */
	float mid;					/**< parameter that influences weight of midrange frequencies */
	float high;					/**< parameter that influences weight of high frequencies */
	FMOD::System *system;		/**< see fmod documentation */
	FMOD::Channel *channel;		/**< see fmod documentation */
	FMOD::Sound *sound;			/**< see fmod documentation */
	int numFreqs;				/**< a duplicate of Config::numFreqs; actually completely useless. */
};

#endif