/**
@file main.cpp
@brief program entry point and GUI definition
*/

#include <QApplication>
#include <QDesktopWidget>
#include <QMainWindow>
#include <QMessageBox>
#include <QMenuBar>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QLabel>
#include "common.h"
#include "MeshViewerWidget.h"
#include <glut.h>

#include <qpushbutton.h>
#include <qslider.h>

void create_menu(QMainWindow &w);
void usage_and_exit(int xcode);

MeshViewerWidget *glWindowWidget;

int main(int argc, char *argv[])
{
	//Qt + OpenMesh
	
	// OpenGL check
  QApplication::setColorSpec( QApplication::CustomColor );
  QApplication app(argc,argv);
#if !defined(__APPLE__)
  glutInit(&argc,argv);
#endif

  if ( !QGLFormat::hasOpenGL() ) {
    QString msg = "System has no OpenGL support!";
    QMessageBox::critical( 0, QString("OpenGL"), msg + QString(argv[1]) );
    return -1;
  }

  int c;
  OpenMesh::IO::Options opt;

  while ( (c=getopt(argc,argv,"hbs"))!=-1 )
  {
     switch(c)
     {
       case 'b': opt += OpenMesh::IO::Options::Binary; break;
       case 'h':
          usage_and_exit(0);
       case 's': opt += OpenMesh::IO::Options::Swap; break;
       default:
          usage_and_exit(1);
     }
  }

  // enable most options for now
  opt += OpenMesh::IO::Options::VertexColor;
  opt += OpenMesh::IO::Options::VertexNormal;
  opt += OpenMesh::IO::Options::VertexTexCoord;
  opt += OpenMesh::IO::Options::FaceColor;
  opt += OpenMesh::IO::Options::FaceNormal;
  opt += OpenMesh::IO::Options::FaceTexCoord;

  // create widget
  QMainWindow mainWin;
  MeshViewerWidget w(&mainWin);
  glWindowWidget = &w;
  glWindowWidget->setOptions(opt);
  //mainWin.setCentralWidget(&w);


  create_menu(mainWin);

  // static mesh, hence use strips
  glWindowWidget->enable_strips();


  //INTERFACE GOES HERE
  //===================
  
  
  QVBoxLayout* vboxTools = new QVBoxLayout();
  vboxTools->setSpacing(50);
  
  //play/stop
  QHBoxLayout* hboxPlayStop = new QHBoxLayout();
  QPushButton* playPauseButton = new QPushButton("Play/Pause"); 
  QPushButton* replayButton = new QPushButton("Restart");
  QObject::connect(playPauseButton,SIGNAL(released()),glWindowWidget,SLOT(playPause()));
  QObject::connect(replayButton,SIGNAL(released()),glWindowWidget,SLOT(rePlay()));
  hboxPlayStop->addWidget(playPauseButton,1,Qt::AlignHCenter);
  hboxPlayStop->addWidget(replayButton,1,Qt::AlignHCenter);
  vboxTools->addLayout(hboxPlayStop,1);


  //number of harmonics
  QHBoxLayout* hboxSmoothing = new QHBoxLayout();
  QLabel* smoothingLabel = new QLabel("Smoothing of movement");
  QSlider* smoothingSlider = new QSlider(Qt::Horizontal);
  smoothingSlider->setMinimum(1);
  smoothingSlider->setMaximum(10);
  smoothingSlider->setValue(2);
  smoothingSlider->setTickPosition(QSlider::TicksAbove);
  smoothingSlider->setTickInterval(1);
  QObject::connect(smoothingSlider,SIGNAL(valueChanged(int)),glWindowWidget,SLOT(changeSmoothing(int)));
  hboxSmoothing->addWidget(smoothingLabel,1,Qt::AlignLeft);
  hboxSmoothing->addWidget(smoothingSlider,1,Qt::AlignLeft);
  vboxTools->addLayout(hboxSmoothing,1);


  //number of harmonics
  QHBoxLayout* hboxNumFreq = new QHBoxLayout();
  QLabel* freqLabel = new QLabel("Number of Harmonics");
  QSlider* freqSlider = new QSlider(Qt::Horizontal);
  freqSlider->setMinimum(6);
  freqSlider->setMaximum(9);
  freqSlider->setValue(6);
  freqSlider->setTickPosition(QSlider::TicksAbove);
  freqSlider->setTickInterval(1);
  QObject::connect(freqSlider,SIGNAL(valueChanged(int)),glWindowWidget,SLOT(changeFreqs(int)));
  hboxNumFreq->addWidget(freqLabel,1,Qt::AlignLeft);
  hboxNumFreq->addWidget(freqSlider,1,Qt::AlignLeft);
  vboxTools->addLayout(hboxNumFreq,1);

  
  //equalizer
  QHBoxLayout* hboxEQ = new QHBoxLayout();
  QSlider* lowSlider = new QSlider(Qt::Vertical);
  QSlider* midSlider = new QSlider(Qt::Vertical);
  QSlider* highSlider = new QSlider(Qt::Vertical);
  lowSlider->setMinimum(0);
  lowSlider->setMaximum(200);
  lowSlider->setValue(100);
  midSlider->setMinimum(0);
  midSlider->setMaximum(200);
  midSlider->setValue(100);
  highSlider->setMinimum(0);
  highSlider->setMaximum(200);
  highSlider->setValue(100);
  QObject::connect(lowSlider,SIGNAL(valueChanged(int)),glWindowWidget,SLOT(eqLow(int)));
  QObject::connect(midSlider,SIGNAL(valueChanged(int)),glWindowWidget,SLOT(eqMid(int)));
  QObject::connect(highSlider,SIGNAL(valueChanged(int)),glWindowWidget,SLOT(eqHigh(int)));
  hboxEQ->addWidget(lowSlider,1,Qt::AlignHCenter);
  hboxEQ->addWidget(midSlider,1,Qt::AlignHCenter);
  hboxEQ->addWidget(highSlider,1,Qt::AlignHCenter);
  vboxTools->addLayout(hboxEQ,1);

  //slider labels
  QHBoxLayout* hboxSliderL = new QHBoxLayout();
  QLabel* lowLabel = new QLabel("low");
  QLabel* midLabel = new QLabel("mid");
  QLabel* highLabel = new QLabel("high");
  hboxSliderL->addWidget(lowLabel,0,Qt::AlignHCenter);
  hboxSliderL->addWidget(midLabel,0,Qt::AlignHCenter);
  hboxSliderL->addWidget(highLabel,0,Qt::AlignHCenter);
  vboxTools->addLayout(hboxSliderL,0);


  //change color
  QPushButton* colorButton = new QPushButton("Change color..."); 
  QObject::connect(colorButton,SIGNAL(released()),glWindowWidget,SLOT(query_change_color()));
  vboxTools->addWidget(colorButton, 1, Qt::AlignHCenter);

  QHBoxLayout* hboxOuter = new QHBoxLayout();
  hboxOuter->addWidget(glWindowWidget,2);
  hboxOuter->addLayout(vboxTools,1);

  //END INTERFACE
  //=============

  QWidget *window = new QWidget();
  window->setLayout(hboxOuter);

  mainWin.setCentralWidget(window);

  mainWin.resize(1024, 768);
  mainWin.show();

  // load scene if specified on the command line
  if ( optind < argc )
  {
	  glWindowWidget->open_mesh_gui(argv[optind]);
  }

  if ( ++optind < argc )
  {
      glWindowWidget->open_audio_gui(argv[optind]);
  }

	
	return app.exec();
}


void create_menu(QMainWindow &w)
{
    using namespace Qt;
    QMenu *fileMenu = w.menuBar()->addMenu(w.tr("&File"));

    QAction* openAct = new QAction(w.tr("&Open mesh..."), &w);
    openAct->setShortcut(w.tr("Ctrl+M"));
    openAct->setStatusTip(w.tr("Open a mesh file"));
    QObject::connect(openAct, SIGNAL(triggered()), glWindowWidget, SLOT(query_open_mesh_file()));
    fileMenu->addAction(openAct);

    QAction* audAct = new QAction(w.tr("Open &audio file..."), &w);
    audAct->setShortcut(w.tr("Ctrl+A"));
    audAct->setStatusTip(w.tr("Open an audio file"));
    QObject::connect(audAct, SIGNAL(triggered()), glWindowWidget, SLOT(query_open_audio_file()));
    fileMenu->addAction(audAct);
}

void usage_and_exit(int xcode)
{
   std::cout << "Usage: meshviewer [-s] [mesh] [texture]\n" << std::endl;
   std::cout << "Options:\n"
	     << "  -b\n"
	     << "    Assume input to be binary.\n\n"
             << "  -s\n"
             << "    Reverse byte order, when reading binary files.\n"
             << std::endl;
   exit(xcode);
}